#include "smooth.h"
#include "vicinity.h"

#include <iostream>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_bspline.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_statistics.h>

void smooth( const std::vector<double> &x, const std::vector<double> &y, std::vector<double> &smoothed_y, int param ) throw ( TError_Generic_error ) {
    smoothed_y.resize(y.size());

    const double tol = 1e-2;

    gsl_vector_const_view gV_x = gsl_vector_const_view_array( &x[0], x.size() );
    gsl_vector_const_view gV_y = gsl_vector_const_view_array( &y[0], y.size() );
    gsl_vector_view       gV_sy = gsl_vector_view_array     ( &smoothed_y[0], smoothed_y.size() );

    if( x.size() < size_t( param*2 ) )
        throw TError_Generic_error( "Not enough elements in vecotr x" );
    {
        size_t n = x.size();
        size_t ncoeffs = n/param;
        size_t nbreak = ncoeffs - 2;
        gsl_bspline_workspace *bw;
        gsl_vector *B;
        gsl_vector *c;
        const gsl_vector *gx, *gy;
        gsl_matrix *X, *cov;
        gsl_multifit_linear_workspace *mw;
        double chisq;//, Rsq, dof, tss;

        /* allocate a cubic bspline workspace (k = 4) */
        bw = gsl_bspline_alloc(4, nbreak);
        B = gsl_vector_alloc(ncoeffs);

        gx = &gV_x.vector;
        gy = &gV_y.vector;

        X = gsl_matrix_alloc(n, ncoeffs);
        c = gsl_vector_alloc(ncoeffs);
        cov = gsl_matrix_alloc(ncoeffs, ncoeffs);
        mw = gsl_multifit_linear_alloc(n, ncoeffs);

        /* use uniform breakpoints on [0, 15] */
        gsl_bspline_knots_uniform(x.front(), x.back(), bw);

        /* construct the fit matrix X */
        for( size_t i = 0; i < n; ++i ) {
            double xi = gsl_vector_get(gx, i);

            /* compute B_j(xi) for all j */
            gsl_bspline_eval(xi, B, bw);

            /* fill in row i of X */
            for( size_t j=0; j < ncoeffs; ++j ) {
                double Bj = gsl_vector_get(B, j);
                gsl_matrix_set(X, i, j, Bj);
            }
        }

        /* do the fit */
        //size_t rank;
        gsl_multifit_linear(X, gy, c, cov, &chisq, mw);

        /*        dof = n - ncoeffs;
                tss = gsl_stats_wtss(w->data, 1, gy->data, 1, gy->size);
                Rsq = 1.0 - chisq / tss;

                fprintf(stderr, "chisq/dof = %e, Rsq = %f\n",
                        chisq / dof, Rsq);
        */
        /* output the smoothed curve */
        {
            double yi, yerr;

            //printf("#m=1,S=0\n");
            for( size_t i=0; i<n; i++ ) {
                gsl_bspline_eval(x[i], B, bw);
                gsl_multifit_linear_est(B, c, cov, &yi, &yerr);
                smoothed_y[i] = yi;
            }
        }

        gsl_bspline_free(bw);
        gsl_vector_free(B);
        gsl_matrix_free(X);
        gsl_vector_free(c);
        gsl_matrix_free(cov);
        gsl_multifit_linear_free(mw);

    }
}
